package hipi.compiler.checker;

import hipi.compiler.*;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;
import java.util.TreeSet;

/**
 * Checks names and types visiting the Abstract Syntax Tree generated by
 * the parser. Names are mapped into integers and the AST is filled with new
 * informations.
 * <h3>Two words on how names are mapped into integers</h3>
 * At the begin of each schedule an empty set of indexes (integer values)
 * is created. Each time a new name is defined, the minimum value not present
 * in the set is associated to it and is inserted.
 * Each time a block, a recursive definition or a thread is visited, a copy of
 * the indexes set is used, so index assignements are optimized because
 * different names in different blocks can be mapped to the same index.
 *
 * @author Paolo Milazzo
 */
public class TreeChecker implements HiPiCompilerVisitor {

	// a reference to the main parser (used to call error handling methods)
	private HiPiCompiler parser;

	// the symbol table used to check names and store temp informations
	private SymbolTable symbolTable = new SymbolTable();
	
	/*
	 * first index used for name mapping, lower numbers are reserved
	 * reservations:
	 *  0  -  functions encoding
	 *  1  -  functions encoding
	 *  2  -  functions encoding
	 *  3  -  SyncToAsinc encoding
	 *  4  -  AddForwarders encoding
	 */
	public static final int ENV_START_IDX = 5;
    
	// chooses a fresh index. argument ts is the set of already used indexes
	public static int chooseIndex(TreeSet ts) {
		Iterator indexIterator = ts.iterator();
		int i = ENV_START_IDX;
		while (indexIterator.hasNext()) {
			if (i<((Integer)indexIterator.next()).intValue()) {
				ts.add(new Integer(i));
				return i;
			}
			else
				i++;
		}
		ts.add(new Integer(i));
		return i;
	}

	/**
	 * Default constructor
	 * 
	 * @param p the parser
	 */
	public TreeChecker(HiPiCompiler p) {
		parser = p; 
	}

	/**
	 * Used to visit the root of the AST.
	 * Creates an empty set of indexes for names mapping, starts a
	 * TreeCheckerFirstVisitor to insert recursive definitions names in
	 * symbol table and continues visit.
	 * 
	 * @param node the root node
	 * @param data is not used
	 * @return null
	 * @see TreeCheckerFirstVisitor
	 * @see SymbolTable
	 */
	public Object visit(ASTHiPiProgram node, Object data) {
		// insert global recDefs names in symbol table
		node.childrenAccept(new TreeCheckerFirstVisitor(parser), 
							new TreeCheckerData(symbolTable,null));
		TreeSet indexSet = new TreeSet();
		node.childrenAccept(this,indexSet);
		return null;
	}

	/**
	 * Used to visit global channel declarations in the AST.
	 * Channel name is inserted in symbol table and channel URI
	 * well form is checked
	 * 
	 * @param node AST node for global channel declaration
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 * @see java.net.URI
	 * @see SymbolTable
	 */
	public Object visit(ASTGlobalChannelDeclaration node, Object data) {
		try {
			// index is not needed for global channels, -1 is used in future
			// visits to recognize a global channel
			symbolTable.addElement(new STE_Channel(node.name, node.params, true, -1));
		} 
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node, ste.getMessage());
		}
		try {
			new URI(node.uri);	// check uri syntax
		} catch (URISyntaxException e) {
			parser.hipi_syntax_error(node, "Malformed URI " + node.uri);
		} 
		node.childrenAccept(this,data); // no children expected
		return null;
	}

	/**
	 * Used to visit recursive definitions in the AST.
	 * Their names were inserted in symbol table by TreeCheckerFirstVisitor,
	 * now are inserted parameters and children (statements inside the
	 * body) are visited.
	 * 
	 * @param node AST node for recursive definition
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 * @see TreeCheckerFirstVisitor
	 * @see SymbolTable
	 */
	public Object visit(ASTRecursiveDefinition node, Object data) {
		// TODO controllare che nome recDef sia diverso da ___thread___*
		symbolTable.addEnv();
		TreeSet indexSet = new TreeSet((TreeSet) data);
		FormalParamDescriptor tmp;
		if (node.params != null) {
			for (int i=0; i<node.params.size(); i++) {
				tmp = (FormalParamDescriptor) node.params.get(i);
				tmp.envIndex = chooseIndex(indexSet);
				try {
					switch (tmp.type) {
						case Types.CHANNEL :
							symbolTable.addElement(
								new STE_Channel(tmp.name,tmp.channelType,false,tmp.envIndex));
							break;
						case Types.INT :
						case Types.STRING :
							symbolTable.addElement(
								new STE_Generic(tmp.name,tmp.type,tmp.envIndex));
							break;
						case Types.TYPEID :
						    // little abuse: channelType contains realType
							int realType=symbolTable.getRealType((String)tmp.channelType.get(0));
								if (realType==Types.CHANNEL)
									// TODO the second param should not be null, type checking is not implemented
									symbolTable.addElement(
										new STE_Channel(tmp.name, null, false, tmp.envIndex));
								else
									symbolTable.addElement(
										new STE_Generic(tmp.name, realType, tmp.envIndex));
							break;
						default : // TODO TYPE CHECKING
							parser.hipi_syntax_error(node, "type error");
					}
				}
				catch (SymbolTableException ste) {
					parser.hipi_syntax_error(node, ste.getMessage());
				}
			}

		}
		node.childrenAccept(this,indexSet);
		
		try {
			symbolTable.removeEnv();
		} 
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node, ste.getMessage());
		}

		return null;
	}
	
	/**
	 * Used to visit schedules in the AST.
	 * Checks colocations, starts a RecDefsFirstVisit to insert
	 * recursive definition names into symbol table and continues
	 * visit.
	 * 
	 * @param node AST node for schedule
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 * @see TreeCheckerFirstVisitor
	 * @see SymbolTable
	 */
	public Object visit(ASTSchedule node, Object data) {
		// checks colocations
		if (node.colocations!=null) {
			String id = null; SymbolTableElement elem;
			for (int i=0; i<node.colocations.size(); i++) {
				id = (String) node.colocations.get(i);
				// TODO: better type checking
				try {
					elem = symbolTable.getElement(id);
					if (!(elem.getType() == Types.CHANNEL)) {
						System.out.println(HiPiCompiler.COMP_DESC + id + " is not a channel");
						System.exit(1);
					}
					if (!((STE_Channel) elem).isGlobal()) {
						System.out.println(HiPiCompiler.COMP_DESC + id + " is not global");
						System.exit(1);
					}
				}
				catch(SymbolTableException ste) {
					parser.hipi_syntax_error(node,ste.getMessage());
				}
			}
		}
	
		symbolTable.addEnv();
		TreeSet indexSet = new TreeSet((TreeSet) data);
		// insert schedule names (recDefs and declarations) in symbol table		
		node.childrenAccept(new TreeCheckerFirstVisitor(parser),
							new TreeCheckerData(symbolTable,indexSet));
		node.childrenAccept(this,indexSet);
		
		try{
			symbolTable.removeEnv();
		}
		catch(SymbolTableException ste) {
			parser.hipi_syntax_error(node, ste.getMessage());
		}
	
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTScheduleMain node, Object data) {
		// TODO check for unique main
		node.childrenAccept(this,data);
		return null;
	}

	
	/**
	 * Used to visit blocks in the AST.
	 * 
	 * @param node AST node for block
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 */
	public Object visit(ASTBlock node, Object data) {

		symbolTable.addEnv();
		TreeSet indexSet = new TreeSet((TreeSet) data);
		node.childrenAccept(this,indexSet);

		try{
			symbolTable.removeEnv();
		}
		catch(SymbolTableException ste) {
			parser.hipi_syntax_error(node, ste.getMessage());
		}
		return null;
	}

	/**
	 * Used to visit local declarations in the AST.
	 * A new index is choosen for name mapping, symbol table is
	 * updated and type is checked.
	 * 
	 * @param node AST node for local declaration
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 * @see SymbolTable
	 */
	public Object visit(ASTLocalDeclarationStatement node, Object data) {		
		if (!node.TC_visited) {
			try {
				node.envIndex = chooseIndex((TreeSet) data);
				// TODO: better type checking
				if (node.type==Types.CHANNEL) 
					symbolTable.addElement(
						new STE_Channel(node.name, node.params, false, node.envIndex));
				else if (node.type==Types.TYPEID) {
					int realType = symbolTable.getRealType(node.typeId);
					if (realType==Types.CHANNEL)
						// TODO the second param should not be null, type checking is not implemented
						symbolTable.addElement(
							new STE_Channel(node.name, null,false, node.envIndex));
					else
						symbolTable.addElement(
							new STE_Generic(node.name, realType, node.envIndex));
				}
				else
					symbolTable.addElement(
						new STE_Generic(node.name, node.type, node.envIndex));

			}
			catch (SymbolTableException ste) {
					parser.hipi_syntax_error(node,ste.getMessage());
			}
		node.childrenAccept(this,data);
		}
		return null;
	}

	/**
	 * Used to visit send operation in the AST.
	 * Checks the type of the channel used for communication and
	 * continues visit.
	 * 
	 * @param node AST node for send
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 */
	public Object visit(ASTSend node, Object data) {
		try {
			// TODO: type checking
			if (symbolTable.getType(node.name) == Types.CHANNEL)
				node.envIndex = 
					((STE_Channel) symbolTable.getElement(node.name)).getEnvIndex();
			else
				node.envIndex = 
					((STE_Generic) symbolTable.getElement(node.name)).getEnvIndex();
		}
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node,ste.getMessage());
		}

		node.childrenAccept(this,data);
		return null;
	}

	/**
	 * Used to visit recv operation in the AST.
	 * Checks type of the channel used for communication, inserts
	 * parameters into symbol table and continues visit.
	 * 
	 * @param node AST node for recv
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 */
	public Object visit(ASTRecv node, Object data) {
		try {
			// TODO: type checking
			if (symbolTable.getType(node.name) == Types.CHANNEL)
				node.envIndex = 
					((STE_Channel) symbolTable.getElement(node.name)).getEnvIndex();
			else
				node.envIndex = 
					((STE_Generic) symbolTable.getElement(node.name)).getEnvIndex();
		}
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node,ste.getMessage());
		}

		// updating the symbol table with parameters received
		FormalParamDescriptor tmp;
		for (int i=0; i<node.params.size(); i++) {
			tmp = (FormalParamDescriptor) node.params.get(i);
			tmp.envIndex = chooseIndex((TreeSet) data);			
			try {
				switch (tmp.type) {
					case Types.CHANNEL :
						symbolTable.addElement(
							new STE_Channel(tmp.name,tmp.channelType,false,tmp.envIndex));
						break;
					case Types.INT :
					case Types.STRING :
						symbolTable.addElement(
							new STE_Generic(tmp.name,tmp.type,tmp.envIndex));
						break;
					case Types.TYPEID :
						// little abuse: channelType contains realType
						int realType=symbolTable.getRealType((String)tmp.channelType.get(0));
							if (realType==Types.CHANNEL)
								// TODO the second param should not be null, type checking is not implemented
								symbolTable.addElement(
									new STE_Channel(tmp.name, null, false, tmp.envIndex));
							else
								symbolTable.addElement(
									new STE_Generic(tmp.name, realType, tmp.envIndex));
						break;
					default : // TODO TYPE CHECKING
						parser.hipi_syntax_error(node, "type error");
				}
			}
			catch (SymbolTableException ste) {
				parser.hipi_syntax_error(node, ste.getMessage());
			}

		}

		node.childrenAccept(this,data);  // no children expected
		return null;
	}

	/**
	 * Used to visit recursive definition calls in the AST.
	 * Checks type of the recursive definition and continues
	 * visit.
	 * 
	 * @param node AST node for recursive definition call
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 */
	public Object visit(ASTRecursiveDefinitionCall node, Object data) {
		try {
			// TODO: type checking
			symbolTable.getType(node.name); 
		}
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node,ste.getMessage());
		}

		node.childrenAccept(this,data);
		return null;
	}

	/**
	 * Used to visit spawn statements in the AST.
	 * Checks type of the destination if needed and continues visit.
	 * 
	 * @param node AST node for spawn
	 * @param data TreeSet containing indexes used for name mapping
	 * @return null
	 */
	public Object visit(ASTSpawnStatement node, Object data) {
		if (node.dest!=null) {
			try {
				// TODO: type checking
				if (symbolTable.getType(node.dest) == Types.CHANNEL)
					node.envIndex = 
						((STE_Channel) symbolTable.getElement(node.dest)).getEnvIndex();
				else
					parser.hipi_syntax_error(node,"type error");
			}
			catch (SymbolTableException ste) {
				parser.hipi_syntax_error(node,ste.getMessage());
			}
		}
		
		symbolTable.addEnv();
		
		node.childrenAccept(this,data);
		
		try {
			symbolTable.removeEnv();
		} 
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node, ste.getMessage());
		}
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTSelectionStatement node, Object data) {
		node.childrenAccept(this,data);
		// TODO implementation
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTExpOp node, Object data) {
		node.childrenAccept(this,data);
		// TODO implementation
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTNewChannel node, Object data) {
		node.childrenAccept(this,data); // no children expected
		// TODO implementation
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTReturnStatement node, Object data) {
		node.childrenAccept(this,data);
		// TODO implementation
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTExpIdentifier node, Object data) {
		try {
			if (symbolTable.getType(node.name) == Types.CHANNEL)
				node.envIndex = 
					((STE_Channel) symbolTable.getElement(node.name)).getEnvIndex();
			else
				node.envIndex = 
					((STE_Generic) symbolTable.getElement(node.name)).getEnvIndex();

		} 
		catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node,ste.getMessage());
		}
		node.childrenAccept(this,data); // no children expected
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTExpInteger node, Object data) {
		node.childrenAccept(this,data); // no children expected
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTExpString node, Object data) {
		node.childrenAccept(this,data); // no children expected
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(SimpleNode node, Object data) {
		parser.hipi_syntax_error(node,"ASTSimpleNode visited");
		node.dump("");
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTTypeDeclaration node, Object data) {
		try {
			symbolTable.addElement(new STE_TypeId(node.name,node.realType));
		} catch (SymbolTableException ste) {
			parser.hipi_syntax_error(node, ste.getMessage());
		}
		return null;
	}

	/**
	 * TODO write comment here
	 */
	public Object visit(ASTImportDeclaration node, Object data) {
		parser.hipi_syntax_error(node,"ASTImportDeclaration visited");
		return null;
	}



}
