/*
 * Simulator.java
 *
 * Created on 24-lug-2003
 */
package hipi.simulator;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Random;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.w3c.dom.Document;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * 
 * @author Paolo Milazzo
 */
public class Simulator extends DefaultHandler {

	public static final URI CONSOLE_STRING_URI = URI.create("sim://string.console");
	public static final URI CONSOLE_INT_URI = URI.create("sim://int.console");
	public static final URI CONSOLE_CHANNEL_URI = URI.create("sim://channel.console");

	// pool of simulated virtual machines
	private Vector VM = new Vector();

	// maps VM names into VMs
	private HashMap VMMap = new HashMap();
	
	// maps URIs into VMs
	// this is used to find location of channels
	private HashMap dns = new HashMap();
	
	// DOM parser for schedule files
	private DocumentBuilder parser;
	
	// simulator console for input
	public static BufferedReader consoleInput = new BufferedReader(new InputStreamReader(System.in));
	
	// level of verbosity (initialize to off) 
	public boolean verbose = false;
	
	// (pseudo)random numbers generator
	private Random rng;
	
	// default constructor
	public Simulator(String netFile, Long seed) {
		
		// parse the network descriptor file
		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser parser = factory.newSAXParser();
			parser.parse(netFile, this);
		} 
		catch (FactoryConfigurationError e) {
			System.err.println("unable to get a SAX parser factory");
			System.exit(1);
		} 
		catch (ParserConfigurationException e) {
			System.err.println("SAX parser was unable to be configured");
			System.exit(1);
		}
		catch (SAXException e) {
			System.err.println("parsing error");
			System.exit(1);
		} 
		catch (IOException e) {
			System.err.println("I/O error");
			System.exit(1);
		}
		
		// initialize the (pseudo)random numbers generator
		if (seed==null)
			rng = new Random();
		else {
			rng = new Random(seed.longValue());
			System.out.println("RNG seed = " + seed);
		}
					
		// initialize the DOM parser for schedules
		try {
			DocumentBuilderFactory factory = 
				DocumentBuilderFactory.newInstance();
			parser = factory.newDocumentBuilder();
		}
		catch (FactoryConfigurationError e) {
			System.err.println("unable to get a document builder factory");
			System.exit(1);
		} 
		catch (ParserConfigurationException e) {
			System.err.println("DOM parser was unable to be configured");
			System.exit(1);
		}
	}

	// executed during netFile parsing each time a tag is encountered
	public void startElement(String uri, String localName, 
					String qname, Attributes atts) 
					throws SAXException {
		
		// if it is <vm name="..."> creates a new virtual machine
		if (qname.equals("vm")) {
			VirtualMachine vm = new VirtualMachine(atts.getValue(0), dns);
			VM.add(vm);
			VMMap.put(atts.getValue(0),vm); 
		}
			
		// if it is <channel uri="..."> creates a new channel
		else if (qname.equals("channel")) {
			try {
				VirtualMachine currentVM = (VirtualMachine)VM.lastElement();
				currentVM.createChannel(new URI(atts.getValue(0)));
			}
			catch(URISyntaxException e) {
				System.out.println(atts.getValue(0) + 
					" is not a valid URI");
			}
		}
	}

	// open a console for schedule loading
	public void openConsole() {
		
		// start the console
		System.out.println("Verbose mode is " + (verbose?"on":"off"));
		System.out.println("Starting console...");
		console_help();
		while (true) {
			System.out.print("SIM> ");
			String command = null;
			try {
				command = consoleInput.readLine();
			} catch (IOException e) {
				System.out.println("I/O error");
			}
			
			// execute command
			if (command!=null) {
				// single step
				if (command.equals("s")) {
					step();
				}
				// multi steps
				else if ((command.length() >= 5 ) && 
						(command.substring(0,4).equals("step"))) {
					try {
						int n = Integer.parseInt(command.substring(5).trim());
						for (int i=0; i<n; i++) { step(); }				
					}
					catch (Exception e) {
						System.out.println("Syntax error");
					}
				}
				// load a schedule
				else if ((command.length() >= 5 ) && 
						(command.substring(0,4).equals("load"))) {
					String params = command.substring(5).trim();
					String VMName = params.substring(0,params.indexOf(" "));
					String schedFile = params.substring(params.indexOf(" ")).trim();
					// parse schedFile and send it to Virtual Machine
					try {
						Document doc = parser.parse(schedFile);
						((VirtualMachine)VMMap.get(VMName)).loadSchedule(doc);
					} 
					catch (SAXException e) {
						System.err.println("schedule parsing error");
					} 
					catch (IOException e) {
						System.err.println("I/O error");
					}
				}
				// infinite steps
				else if (command.equals("play")) {
					while(true) { step(); }
				}
				// print actual configuration
				else if (command.equals("view")) {
					for (int i=0; i<VM.size(); i++) 
						((VirtualMachine) VM.get(i)).view();
				}
				// help message
				else if (command.equals("help")) {
					console_help();
				}
				// exit the simulator
				else if (command.equals("quit")) {
					System.out.println("Bye...");
					System.exit(0);
				}
				// set verbose
				else if (command.equals("verb")) {
					verbose = !verbose;
					System.out.println("Verbose mode is " + (verbose?"on":"off"));
				}
				else if (command.equals("")) { /* do nothing */ }
				// command not found
				else {
					System.out.println("Command not found");
					System.out.println("Type \"help\" for more info");
				}
			}
		}
	}

	// prints console help message
	private void console_help() {
		System.out.println("Commands:");
		System.out.println("  load <VM_name> <filename>   - Load a schedule on a VM");
		System.out.println("  play                        - Execute forever");
		System.out.println("  step <num>                  - Execute <num> steps");
		System.out.println("  s                           - The same as: step 1");
		System.out.println("  view                        - Print simulator state");
		System.out.println("  help                        - This message");
		System.out.println("  verb                        - Switch verbose mode on|off");
		System.out.println("  quit                        - Exit the simulator");
		System.out.println();
		System.out.println("ATT: Console is case sensitive");
		System.out.println();		
	}

	// executes one simulation step
	private void step() {
		// executes one step on a randomly choosen VirtualMachine
		((VirtualMachine)VM.get(rng.nextInt(VM.size()))).step(verbose);
		System.out.flush();
	}
		
	// main function
	public static void main(String[] args) {
		Simulator sim = null;
		if (args.length == 1) {
			sim = new Simulator(args[0],null);
		}
		else if (args.length == 3){
			if (!(args[0].equals("-s") || args.equals("--seed")))
				print_usage();
			else {
				try {
					sim = new Simulator(args[2],new Long(args[1]));
				}
				catch (Exception e) {
					System.out.println("Option value not valid");
					print_usage();
					System.exit(1);
				}
			}
		}
		else {
			print_usage();
			System.exit(1);
		}
		System.out.println("Simulator initialized...");
		sim.openConsole();
	}
	
	// prints usage message
	private static void print_usage() {
		System.out.println("Usage: sim [options] <net_descriptor_filename>");
		System.out.println();
		System.out.println("Options:");
		System.out.println("   -s <seed> | --seed <seed>");
		System.out.println("   Set the seed of the random numbers generator");
		System.out.println();		
	}
	
}
